/*
**  cfg_vers.c -- Version Information for OSSP cfg (syntax: C/C++)
**  [automatically generated and maintained by GNU shtool]
*/

#ifdef _CFG_VERS_C_AS_HEADER_

#ifndef _CFG_VERS_C_
#define _CFG_VERS_C_

#define __CFG_VERSION 0x009203

typedef struct {
    const int   v_hex;
    const char *v_short;
    const char *v_long;
    const char *v_tex;
    const char *v_gnu;
    const char *v_web;
    const char *v_sccs;
    const char *v_rcs;
} __cfg_version_t;

extern __cfg_version_t __cfg_version;

#endif /* _CFG_VERS_C_ */

#else /* _CFG_VERS_C_AS_HEADER_ */

#define _CFG_VERS_C_AS_HEADER_
#include "cfg_vers.c"
#undef  _CFG_VERS_C_AS_HEADER_

__cfg_version_t __cfg_version = {
    0x009203,
    "0.9.3",
    "0.9.3 (23-Apr-2004)",
    "This is OSSP cfg, Version 0.9.3 (23-Apr-2004)",
    "OSSP cfg 0.9.3 (23-Apr-2004)",
    "OSSP cfg/0.9.3",
    "@(#)OSSP cfg 0.9.3 (23-Apr-2004)",
    "$Id: cfg_vers.c,v 1.6 2004/04/23 10:28:01 thl Exp $"
};

#endif /* _CFG_VERS_C_AS_HEADER_ */

